#include <iostream>
#include <unistd.h>

#include "BinaryFuncBase.h"

BinaryFuncBase::BinaryFuncBase() :
    m_module_name(""),
    m_pBinarySocket(nullptr),
    m_pBinaryThreadPool(nullptr)
{
    Pri_AddNewGenericCommand( "GetModuleName",   static_cast<CommandFunc>(&BinaryFuncBase::Pri_GetModuleName),    0 );
}

BinaryFuncBase::~BinaryFuncBase()
{
    m_pBinaryThreadPool.reset();
    m_pBinarySocket.reset();
    m_generic_func_table.clear();
    m_func_table.clear();
    m_module_name = "";
}

void BinaryFuncBase::Initialize(std::string i_module_name)
{
    m_module_name = i_module_name;

    m_pBinarySocket = std::unique_ptr<BinarySocket>(new BinarySocket());
    m_pBinarySocket->Initialize();
    m_pBinarySocket->RegisterCallbackFunc(std::bind(&BinaryFuncBase::Pri_ReceiveCommand, this, std::placeholders::_1));
    m_pBinarySocket->StartReceive();

    m_pBinaryThreadPool = std::unique_ptr<BinaryThreadPool>(new BinaryThreadPool());
    m_pBinaryThreadPool->RegisterCallbackFunc(std::bind(&BinaryFuncBase::Pri_SendResultMessage, this, std::placeholders::_1, std::placeholders::_2, std::placeholders::_3, std::placeholders::_4));
}

void BinaryFuncBase::AddNewCommand(std::string i_command, CommandFunc i_func, int i_arg_num)
{
    std::pair<std::string, FuncTableItem> tmp(i_command, {i_func, i_arg_num});
    m_func_table.insert(tmp);
}

void BinaryFuncBase::Pri_AddNewGenericCommand(std::string i_command, CommandFunc i_func, int i_arg_num)
{
    std::pair<std::string, FuncTableItem> tmp(i_command, {i_func, i_arg_num});
    m_generic_func_table.insert(tmp);
}

BinaryFuncBaseRet BinaryFuncBase::Pri_ParseCommand(std::string i_command, FuncItem &o_func_item, std::vector<std::string> &o_ret_msg)
{
    // std::cout << "@@@@@@@@@@" << __FILE__ << " " << __FUNCTION__ << std::endl;

    std::vector<std::string> command_array;
    std::unordered_map<std::string, FuncTableItem>::iterator it_table;
    std::vector<std::string> input_arg;

    Pri_StringSplit(i_command, ' ', command_array);
    if (!command_array.empty())
    {
        if (command_array[0].compare("GENERIC") == 0)
        {
            o_func_item.func_type = BINARY_FUNC_GENERIC;
            o_func_item.command = command_array[1];
            it_table = m_generic_func_table.find(o_func_item.command);
            if (it_table == m_generic_func_table.end() || it_table->second.func == nullptr)
            {
                o_ret_msg.push_back("Generic command not found");
                return BINARY_FUNC_BASE_RET_NG;
            }
        }
        else
        {
            o_func_item.func_type = BINARY_FUNC_UNIQUE;
            o_func_item.command = command_array[1];
            it_table = m_func_table.find(o_func_item.command);
            if (it_table == m_func_table.end() || it_table->second.func == nullptr)
            {
                o_ret_msg.push_back("Command not found");
                return BINARY_FUNC_BASE_RET_NG;
            }
        }

        std::string tmp_string;
        command_array[2].erase(command_array[2].begin(), command_array[2].begin() + 4);
        command_array[2].erase(command_array[2].end());
        o_func_item.id = std::stoul(command_array[2].c_str());
        if (command_array.size() > 2)
        {
            auto it = command_array.begin() + 3;
            it->erase(it->begin(), it->begin() + 5);
            it->erase(it->end());
            tmp_string = *it;
        }

        if (it_table->second.arg_num != 0)
        {
            if (it_table->second.arg_num > 1)
            {
                Pri_StringSplit(tmp_string, ' ', input_arg);
            }
            else
            {
                input_arg.push_back(tmp_string);
            }

            if (it_table->second.arg_num != (int)input_arg.size())
            {
                o_ret_msg.push_back("Invalid argument number");
                return BINARY_FUNC_BASE_RET_NG;
            }
        }

        o_func_item.func = std::bind(static_cast<CommandFunc>(it_table->second.func), this, std::placeholders::_1, std::placeholders::_2);
        o_func_item.input_arg = input_arg;
    }
    else
    {
        o_ret_msg.push_back("Invalid command");
        return BINARY_FUNC_BASE_RET_NG;
    }

    return BINARY_FUNC_BASE_RET_OK;
}

void BinaryFuncBase::Pri_ReceiveCommand(std::string i_command)
{
    // std::cout << "@@@@@@@@@@" << __FILE__ << " " << __FUNCTION__ << std::endl;

    std::string command(i_command);
    FuncItem func_item;
    std::vector<std::string> ret_msg;

    if (i_command.find("GENERIC") == std::string::npos)
    {
        Pri_SendReceiveMessage(command);
    }

    if (Pri_ParseCommand(command, func_item, ret_msg) != BINARY_FUNC_BASE_RET_OK)
    {
        if (func_item.func_type == BINARY_FUNC_UNIQUE)
        {
            Pri_SendResultMessage(COMMAND_RET_NG, func_item.command, func_item.id, ret_msg);
            return;
        }
        else
        {
            Pri_SendResultMessage(COMMAND_RET_GENERIC_NG, func_item.command, func_item.id, ret_msg);
            return;
        }
    }

    Pri_PushTaskToThread(func_item);
}

void BinaryFuncBase::Pri_PushTaskToThread(FuncItem i_func_item)
{
    // std::cout << "@@@@@@@@@@" << __FILE__ << " " << __FUNCTION__ << std::endl;

    m_pBinaryThreadPool->AssignTask(i_func_item);
    // std::cout << "~~~~~~~~~~" << __FILE__ << " AssignTask complete" << std::endl;
}

void BinaryFuncBase::Pri_SendReceiveMessage(std::string i_command)
{
    // std::cout << "@@@@@@@@@@" << __FILE__ << " " << __FUNCTION__ << std::endl;

    std::string reply_message = i_command + " \"RESULT=RECEIVE\"";

    Pri_SendReplyMessage(reply_message);
}

void BinaryFuncBase::Pri_SendResultMessage(CommandRet i_enum, std::string i_command, int i_id, std::vector<std::string> i_value)
{
    // std::cout << "@@@@@@@@@@" << __FILE__ << " " << __FUNCTION__ << std::endl;

    std::string reply_message;

    switch (i_enum)
    {
        case COMMAND_RET_OK:
            reply_message = m_module_name + " " + i_command + " \"ID=" + std::to_string(i_id) + "\"" + " \"RESULT=OK\"";
            reply_message += " \"MSG=";
            if (!i_value.empty())
            {
                for (auto it = i_value.begin(); it != i_value.end(); it ++)
                {
                    reply_message += it->c_str();
                }
            }
            reply_message += "\"";
            break;

        case COMMAND_RET_NG:
            reply_message = m_module_name + " " + i_command + " \"ID=" + std::to_string(i_id) + "\"" + " \"RESULT=NG\"";
            reply_message += " \"MSG=";
            if (!i_value.empty())
            {
                for (auto it = i_value.begin(); it != i_value.end(); it ++)
                {
                    reply_message += it->c_str();
                }
            }
            reply_message += "\"";
            break;

        case COMMAND_RET_GENERIC_OK:
            reply_message = "GENERIC " + i_command + " \"ID=" + std::to_string(i_id) + "\"" + " \"RESULT=OK\"";
            reply_message += " \"MSG=";
            if (!i_value.empty())
            {
                for (auto it = i_value.begin(); it != i_value.end(); it ++)
                {
                    reply_message += it->c_str();
                }
            }
            reply_message += "\"";
            break;

        case COMMAND_RET_GENERIC_NG:
            reply_message = "GENERIC " + i_command + " \"ID=" + std::to_string(i_id) + "\"" + " \"RESULT=NG\"";
            reply_message += " \"MSG=";
            if (!i_value.empty())
            {
                for (auto it = i_value.begin(); it != i_value.end(); it ++)
                {
                    reply_message += it->c_str();
                }
            }
            reply_message += "\"";
            break;
    }

    Pri_SendReplyMessage(reply_message);
}

void BinaryFuncBase::Pri_SendReplyMessage(std::string i_msg)
{
    // std::cout << "@@@@@@@@@@" << __FILE__ << " " << __FUNCTION__ << std::endl;

    if (!i_msg.empty())
    {
        if (m_pBinarySocket->SendReplyMessage(i_msg) != BINARY_SOCKET_RET_OK)
        {
            std::cout << "Failed to send reply message" << std::endl;
        }
    }
    else
    {
        std::cout << "Reply message is empty" << std::endl;
    }
}

BinaryFuncBaseRet BinaryFuncBase::Pri_StringSplit(std::string i_string, char i_delimiter, std::vector<std::string> &o_string_array)
{
    int count = 0;
    size_t start = 0;
    size_t end = i_string.find(i_delimiter);

    if (i_string.empty() ||
        i_string.find(i_delimiter) == std::string::npos ||
        i_string.at(0) == i_delimiter)
    {
        return BINARY_FUNC_BASE_RET_NG;
    }

    while (end != std::string::npos)
    {
        o_string_array.push_back(i_string.substr(start, end - start));
        start = end + 1;
        end = i_string.find(i_delimiter, start);
        ++count;
    }

    std::string string_stray = i_string.substr(start, end - start);
    if (!string_stray.empty())
    {
        o_string_array.push_back(string_stray);
    }

    return BINARY_FUNC_BASE_RET_OK;
}

FuncRet BinaryFuncBase::Pri_GetModuleName(std::vector<std::string> i_arg, std::vector<std::string> &o_msg)
{
    o_msg.push_back(m_module_name);

    return BINARY_FUNC_RET_OK;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
BinaryThreadPool::BinaryThreadPool() :
    m_pthread(nullptr),
    m_stop_task(false),
    m_function(nullptr)
{
    m_thread_pool.push_back(std::unique_ptr<BinaryThread>(new BinaryThread(1)));
    // m_thread_pool.push_back(std::unique_ptr<BinaryThread>(new BinaryThread(2)));
    // m_thread_pool.push_back(std::unique_ptr<BinaryThread>(new BinaryThread(3)));
    m_pthread = std::unique_ptr<std::thread>(new std::thread(&BinaryThreadPool::Pri_TaskThread, this));
    for (auto it = m_thread_pool.begin(); it != m_thread_pool.end(); it++)
    {
        (*it)->RegisterCallbackFunc(std::bind(&BinaryThreadPool::Pri_GetFuncReturn, this, std::placeholders::_1, std::placeholders::_2, std::placeholders::_3));
    }
}

BinaryThreadPool::~BinaryThreadPool()
{
    Pri_StopTask();
    if (m_pthread)
        m_pthread->join();

    for (auto it = m_thread_pool.begin(); it != m_thread_pool.end(); it++)
    {
        (*it).reset();
    }
    m_thread_pool.clear();
}

void BinaryThreadPool::AssignTask(FuncItem i_func_item)
{
    // std::cout << "&&&&&&&&&&" << __FILE__ << " " << __FUNCTION__ << std::endl;

    std::unique_lock<std::mutex> lck(m_mutex);
    m_func_list.push_back(i_func_item);
    m_cv.notify_one();
}

void BinaryThreadPool::RegisterCallbackFunc(std::function<void(CommandRet, std::string, int, std::vector<std::string>)> i_function)
{
    m_function = i_function;
}

void BinaryThreadPool::Pri_GetFuncReturn(FuncRet i_ret, FuncItem i_func_item, std::vector<std::string> i_msg)
{
    if (i_func_item.func_type != BINARY_FUNC_GENERIC)
    {
        if (i_ret != BINARY_FUNC_RET_OK)
        {
            m_function(COMMAND_RET_NG, i_func_item.command, i_func_item.id, i_msg);
        }
        else
        {
            m_function(COMMAND_RET_OK, i_func_item.command, i_func_item.id, i_msg);
        }
    }
    else
    {
        if (i_ret != BINARY_FUNC_RET_OK)
        {
            m_function(COMMAND_RET_GENERIC_NG, i_func_item.command, i_func_item.id, i_msg);
        }
        else
        {
            m_function(COMMAND_RET_GENERIC_OK, i_func_item.command, i_func_item.id, i_msg);
        }
    }
}

bool BinaryThreadPool::Pri_IsTaskPending()
{
    if (m_func_list.empty())
    {
        return false;
    }
    return true;
}

void BinaryThreadPool::Pri_StopTask()
{
    m_stop_task = true;
    m_cv.notify_one();
}

void BinaryThreadPool::Pri_TaskThread()
{
    while(true)
    {
        // std::cout << "&&&&&&&&&&" << __FILE__ << " " << __FUNCTION__ << std::endl;

        std::unique_lock<std::mutex> lck(m_mutex);
        m_cv.wait_for(lck, std::chrono::milliseconds(500), std::bind(&BinaryThreadPool::Pri_IsTaskPending, this));

        if (m_stop_task)
            break;

        for (auto it = m_thread_pool.begin(); it != m_thread_pool.end(); it++)
        {
            if (m_func_list.empty())
                break;

            if ((*it)->IsThreadBusy())
            {
                std::cout << "Thread is busy" << __FILE__ << " " << __FUNCTION__ << std::endl;
                continue;
            }

            auto current_func_data = m_func_list.front();
            if ((*it)->AssignTask(current_func_data))
            {
                m_func_list.erase(m_func_list.begin());
            }
        }
        lck.unlock();
        usleep(10*1000);
        m_cv.notify_one();
    }
}